#! python
"""
script: aprcompiler
Author: Ernesto Posse
Created on: July 4, 2006
Last update: Mar 168, 2007

Description:
    This script compiles a .apr file into a python representation.
    If the file name is of the form <some_name>.apr then the following will be 
    generated in the same directory where the file is located:
    
    - <some_name>.py:   A file containing a python representation of the grammar,
    - <some_name>_cfg:  A directory containing a python package with that name 
                        and which contains the following files:
        - __init__.py: .................... A file to identify the <some_name>_cfg 
                                            directory as a python package.
        - <some_name>_cfg_ll1_actions.py: . A python module containing the source 
                                            code for the actions associated with 
                                            grammar rules.
        - <some_name>_cfg_ll1_wan.py: ..... A python module containing the normalized 
                                            form of the python representation of the 
                                            grammar.
        - <some_name>.pkl: ................ A python pickled file containing the parser 
                                            object itself (including the lexer and the 
                                            CFG object.)
    
Typical usage:
    1) Create a source apr file, for example: mylang.apr
    
        separators
            colon ":"
        identifiers
            id
        numbers 
            age
        strings
            name
        start 
            RECORD
        rules
        RECORD -> name colon age : "('by name', $1, $3)"
                | id colon age : "('by id', $1, $3)"
    
    2) On the command line type:
        
        apr mylang.apr
    
    3) To use the generated parser in a python module, do:
    
        from aperiot.parsergen import load_parser
        parser = load_parser('mylang')
        text = '"Ernesto Posse":1329'
        outcome = parser.parse(text)
        name, age = outcome

    Note: the generated files, as well as the aperiot package, must be in the python path (i.e. must appear in sys.path.)
"""

import logging
import os.path
import pprint

from libutilitaspy.general.utils import get_cmd_line_args, fit, indent

import aperiot

from aperiot.parsergen import build_parser, load_parser, save_parser, ParserGenException
from aperiot.llparser import ParsingException
from aperiot.cfgs import CFG, CFGException
#from aperiot.metacfg_wat_cfg import *

logger = logging.getLogger('aperiot')
pp = pprint.PrettyPrinter(indent=4)

__all__ = ['main', 'compile_cfg']

VERSION = aperiot.VERSION

ASCII      = 0
BINARY     = 2
OLD_BINARY = 1

COMMAND_LINE_OPTIONS = \
    {
        ("-p", "--pklformat"): { 
            'nargs':    1,
            'action':   'store', 
            'type':     int, 
            'dest':     'pickle_format', 
            'default':  ASCII,
            'help':     "Pickle format for the generated pkl file; 0 = ASCII, 1 = Old binary, 2 = New binary; The default is 0" 
        },
        ("-f", "--fullsave"): {
            'action':   'store_true',
            'dest':     'full_save',
            'help':     "Full save: saves the generated parser and actions. Necessary when using load_parser instead of build_parser."
        },
        ("-d", "--dir"): {
            'nargs':    1,
            'action':   'store',
            'type':     str,
            'dest':     'target_dir',
            'default':  [os.getcwd()],
            'help':     "Target directory. The default is the current directory (%s)." % os.getcwd()
        },
        ('-r', '--hard'): {
            'action':   'store_false',
            'dest':     'using_bootstrapped',
            'help':     "Use the hard-coded meta-grammar instead of the bootstrapped meta-grammar."
        },
        ('-V', '--version'): {
            'action':   'version',
            'version':  '%(prog)s ' + str(VERSION)
        },
        'source': {
            'help': "The source grammar file."
        }
    }


def compile_cfg(parser, target_dir, source, pkl_format, full_save, using_bootstrapped=True):
    with file(source, 'r') as text:
        logger.info("Parsing source cfg file '"+source+"'")
        grammar = parser.parse(text)
        logger.info("Source cfg file parsed")
        logger.info(pp.pformat(grammar))
    generate_pyrepr(target_dir, source, grammar, using_bootstrapped)
    if full_save:
        imports = grammar[0]
        cfg_description_module, package_name, dir = get_module_pkginfo(source)
        logger.info("Building parser")
        new_parser = build_parser(cfg_description_module)
        logger.info("Parser built")
        logger.info("Saving generated parser")
        save_parser(new_parser, target_dir, cfg_description_module, pickle_format=pkl_format)
        logger.info("Generated parser saved")

def get_module_pkginfo(full_source_name):
    dir, source = os.path.split(full_source_name)
    dir = os.path.abspath(dir)
    cfg_description_module, ext = os.path.splitext(source)
    #~ base_dir, package_name = os.path.split(dir)
    package_name = cfg_description_module + '_cfg'
    if package_name == '':
        package_name = '__main__'
    return cfg_description_module, package_name, dir

def generate_pyrepr(dir, source, grammar, using_bootstrapped=True):
    new_file_name = make_new_file_name(source)
    logger.info("Generating python representation")
    generated_text = generate_text(grammar, using_bootstrapped)
    logger.info("Python representation generated")
    cfg_description_module, package_name, dir = get_module_pkginfo(source)
    package_path = CFG.get_package_path(package_name)
    if package_path is None:
        logger.info("Creating new package: '" + package_name + "' in '" + str(os.getcwd()) + "'") 
        package_path = CFG.create_package(package_name, dir)
    else:
        abs_package_path = os.path.abspath(package_path)
        logger.info("Package '" + abs_package_path + "' already exists. Saving new modules there.")
    new_file_name = os.path.join(package_path, new_file_name)
    new_file = file(new_file_name, 'w')
    new_file.write(generated_text)
    new_file.close()
    abs_new_file_name = os.path.abspath(new_file_name)
    logger.info("Python representation saved in file '"+abs_new_file_name+"'")

def extract_terminals(terminals):
    try:
        keywords    = terminals['keywords']
    except KeyError:
        keywords    = []
    try:
        operators   = terminals['operators']
    except KeyError:
        operators   = []
    try:
        separators  = terminals['separators']
    except KeyError:
        separators  = []
    try:
        brackets    = terminals['brackets']
    except KeyError:
        brackets    = []
    try:
        identifiers = terminals['identifiers']
    except KeyError:
        identifiers = []
    try:
        numbers     = terminals['numbers']
    except KeyError:
        numbers     = []
    try:
        strings     = terminals['strings']
    except KeyError:
        strings     = []
    return (keywords, operators, separators, brackets, identifiers, numbers, strings)
    
def generate_text(grammar, using_bootstrapped=True):
    logger.info("using bootstrapped is " + str(using_bootstrapped))
    if using_bootstrapped:
        imports, lexer, useindents, usenewlines, terminals, start, rules = grammar
    else:
        imports, useindents, usenewlines, terminals, start, rules = grammar
    keywords, operators, separators, brackets, identifiers, numbers, strings = extract_terminals(terminals)
    rules_text = generate_rules_text(rules, terminals)
    generated_text = "\n"
    generated_text += "from aperiot.lexer import *\n\n"
    generated_text += "imports = [" + ', '.join(["'" + item + "'" for item in imports]) + "]\n"
    generated_text += "\n"
    if using_bootstrapped and lexer:
        generated_text += "lexer = '"+str(lexer)+"'\n"
        generated_text += "\n"
    generated_text += "useindents = "+str(useindents)+"\n"
    generated_text += "usenewlines = "+str(usenewlines)+"\n"
    generated_text += "\n"
    generated_text += "# Terminal symbols\n\n"
    generated_text += generate_terminal_defs(useindents, usenewlines, keywords, 
                                                operators, separators, brackets, 
                                                identifiers, numbers, strings) 
    generated_text += "# Non-terminal symbols\n\n"
    generated_text += generate_nonterminal_defs(rules)
    generated_text += "# Start symbol\n\n"
    generated_text += "start = '" + start + "'\n\n"
    generated_text += "# Rules\n\n"
    generated_text += "rules = \\\n" + indent(rules_text, 4) + "\n"
    generated_text += "\n"
    return generated_text

def generate_terminal_defs(useindents, usenewlines, keywords, 
                            operators, separators, brackets, 
                            identifiers, numbers, strings):
    terminal_defs = ""
    if useindents:
        terminal_defs += "t_indent  = Indent\n"
        terminal_defs += "t_dedent  = Dedent\n"
    if usenewlines:
        terminal_defs += "t_newline = Newline\n"
    if len(keywords) > 0:
        keyword_strings = [k.string for k in keywords]
        size1 = max(len(k) for k in keyword_strings)
        keyword_symbols = [k.symbolic_name for k in keywords]
        size2 = max(len(k) for k in keyword_symbols)
        size = max(size1, size2)
        for keyword in keywords:
            terminal_defs += 't_' + fit(keyword.symbolic_name, size) + " = " + repr(keyword) + "\n"
    if len(operators) > 0:
        operator_strings = [o.symbolic_name for o in operators]
        size = max(len(o) for o in operator_strings)
        for operator in operators:
            terminal_defs += 't_' + fit(operator.symbolic_name, size) + " = " + repr(operator) + "\n"
    if len(separators) > 0:
        separator_strings = [o.symbolic_name for o in separators]
        size = max(len(o) for o in separator_strings)
        for separator in separators:
            terminal_defs += 't_' + fit(separator.symbolic_name, size) + " = " + repr(separator) + "\n"
    if len(brackets) > 0:
        bracket_strings = [o.symbolic_name for o in brackets]
        size = max(len(o) for o in bracket_strings)
        for bracket in brackets:
            terminal_defs += 't_' + fit(bracket.symbolic_name, size) + " = " + repr(bracket) + "\n"
    if len(identifiers) > 0:
        identifiers_strings = [k.string for k in identifiers]
        size = max(len(k) for k in identifiers_strings)
        for identifier in identifiers:
            terminal_defs += 't_' + fit(identifier.string, size) + " = Identifier\n"
    if len(numbers) > 0:
        number_strings = [o.string for o in numbers]
        size = max(len(o) for o in number_strings)
        for number in numbers:
            terminal_defs += 't_' + fit(number.string, size) + " = NumberLiteral\n"
    if len(strings) > 0:
        string_strings = [o.string for o in strings]
        size = max(len(o) for o in string_strings)
        for string in strings:
            terminal_defs += 't_' + fit(string.string, size) + " = StringLiteral\n"
    terminal_defs += "\n\n"
    return terminal_defs 

def terminal_names(terminals):
    l = ['newline','indent','dedent']
    for kind in terminals:
        for terminal in terminals[kind]:
            l.append(terminal.symbolic_name)
    return l

def generate_nonterminal_defs(rules):
    nonterminal_defs = ""
    size = max(len(key) for key in rules) 
    for key in rules:
        nonterminal_defs += 'n_' + fit(key, size) + " = '" + key + "'\n"
    nonterminal_defs += "\n\n"
    return nonterminal_defs 
    
def generate_rules_text(rules, terminals):
    terminal_symbols = terminal_names(terminals)
    rules_text = "{\n"
    for key in rules:
        rules_text += " " * 4 + 'n_' + key + ": \n"
        rules_text += " " * 8 + "[\n"
        for word, action in rules[key]:
            l = []
            for symbol in word:
                if symbol in terminal_symbols:
                    prefix = 't_'
                else:
                    prefix = 'n_'
                l.append(prefix + str(symbol))
            symbols = ', '.join(l)
            if action == '':
                action = 'None'
            rules_text += " " * 12 + "([" + symbols + "], '" + str(action) + "'), \n"
        rules_text += " " * 8 + "],\n"
    rules_text += "}\n"
    return rules_text

def make_new_file_name(full_source_name, overwrite=True):
    base_path, source = os.path.split(full_source_name)
    root, ext = os.path.splitext(source)
    new_name = root + '.py'
    if not overwrite and os.path.exists(new_name):
        raise CFGCompilerException("The file '" + new_name + "' already exists. Please use a different name for your cfg file, or remove the .py file")
    return new_name

def main():
    try:
        args = get_cmd_line_args(COMMAND_LINE_OPTIONS, description="aperiot parser generator for Python.")
        pickle_format   = args.pickle_format
        target_dir      = os.path.abspath(args.target_dir[0]) 
        full_save       = args.full_save
        using_bootstrapped = args.using_bootstrapped
        if using_bootstrapped:
            meta_cfg_parser = 'metacfg_bootstrap'
        else:
            meta_cfg_parser = 'metacfg_wat'
        full_source_name = args.source
        source = os.path.abspath(full_source_name)
        logger.info("Loading CFG parser")
        parser = load_parser(meta_cfg_parser)
        logger.info("CFG parser loaded")
        compile_cfg(parser, target_dir, source, pickle_format, full_save, using_bootstrapped)
        logger.info("CFG compilation completed")
    except IOError, e:
        logger.error('(IOError) '+str(e))
    except ParsingException, e:
        logger.error('(ParsingException) '+str(e))
    except CFGException, e:
        logger.error('(CFGException) '+str(e))
    except CFGCompilerException, e:
        logger.error('(CompilerException) '+str(e))
    except ParserGenException, e:
        logger.error('(ParserGenerationException) '+str(e))

class CFGCompilerException(Exception):
    pass

if __name__ == "__main__":
    main()
