"""
Module: llparser
Author: Ernesto Posse
Created on: April 13, 2006
Last update: Mar 16, 2007

Description:
  Generic LL(1) parser

Typical use:
    See file 'parspattern.py'

Classes:
    Parser
        Attributes:
            parse_tree: ParseTreeNode       The last tree computed
            cfg: CFG                        The Context Free Grammar of the parser
            table: ParsingTable             The parsing table corresponging to the CFG
        Methods:
            __init__(self: Parser, cfg: CFG, dicard_choices:bool=True)
                Constructor. Creates a new parser from the given CFG. 
                If discard_choices is True, the resulting parsing table is LL(1), that
                is, all its elements are single rule numbers. Otherwise the elements
                of the table are sets of applicable rule numbers. 
                (Note: The parser currently does not support parsing for LL(k) tables 
                with k>1)
            parse(self: Parser, 
                  input_stream: Token iterator, 
                  apply_actions:bool=True, 
                  pass_all:bool=True) -> 'a
                Parses the input stream of tokens into a concrete syntax tree. If the
                apply_actions parameter is True, it applies the actions associated with 
                rules in the CFG to the tree. The outcome of applying these actions is 
                returned. 
                    If pass_all is True, each action is passed as arguments all outcomes 
                associated with both terminal and non-terminal symbols in the RHS of the 
                rule. These arguments are passed in order from left-to-right as they 
                appear in the RHS of the rule. If pass_all is False, the action receives 
                as input only the outcomes associated with non-terminals in the RHS.
                    The input_stream can be any object that supports the iterator protocol.
                In particular it must have an __iter__ method which must return an
                iterator object, i.e. an object with a next method. Normally this 
                would be a list or a generator. 
                    The objects contained in the input stream must by instances of the 
                Token class. The outcome of the Lexer.tokenize method satisfies these 
                conditions.
            get_parse_tree(self: Parser, input_stream: Token iterator) -> ParseTreeNode
                This is the core algorithm for the LL(1) parser. The parameters are
                as in the parse method above. Its outcome is the root node of the
                Concrete Syntax Tree (CST).
            get_derivation(self: Parser, 
                           input_stream: Token iterator, 
                           pass_all:bool=True) -> int list
                This is like the get_parse_tree method, but instead of returning the CST,
                its outcome is the left-most derivation of the input_stream given as a 
                list of CFG rule numbers, appearing in the order in which the rules are 
                applied.
            build_parse_tree(self: Parser, 
                             rule_sequence: int list, 
                             include_terminals:bool=True) -> ParseTreeNode
                This method takes a list of CFG rule numbers (the rule_sequence) and
                returns a pointer to the root of the concrete syntax parse tree.
            apply_actions(self: Parser, node: ParseTreeNode, pass_all:bool=True) -> 'a
                This method applies recursively the actions associated with the rules
                to the concrete syntax tree given (starting from node, doing a 
                depth-first traversal) It returns the outcome of these actions.

    ParsingTable
    ParseTreeNode
    ParsingException
"""

import copy
import logging
import types

from libutilitaspy.data_structures.stacks import Stack
from libutilitaspy.general.utils import short_class_name

import aperiot
from aperiot.cfgs import Terminal, NonTerminal, EmptyString, RuleAction, SymbolSet
from aperiot.lexer import EndOfInput

log = logging.getLogger(__name__)



__all__ = ['Parser', 'ParsingException']


class Parser:
    
    def __init__(self, 
                    cfg, 
                    lexer=None, 
                    rule_actions_module=None, 
                    discard_choices=True):
        self.table = ParsingTable(cfg, discard_choices)
        self.cfg = cfg
        self.lexer = lexer
        self.rule_actions_module = rule_actions_module
        self.token_list = []
        self.modules_to_import = self.cfg.modules_to_import
    
    """
    Public instance methods
    """
    
    # Public accessors, selectors and extractors
    
    def __getstate__(self):
        #~ return {'dummy_parser':'dummy'}
        obj_dict = copy.copy(self.__dict__)
        if type(self.rule_actions_module) is types.ModuleType:
            obj_dict['rule_actions_module'] = self.rule_actions_module.__name__
        return obj_dict

    def import_action_modules(self):
        """Import modules used by the grammar's action"""
        for module in self.modules_to_import:
            exec "from " + module + " import *" in self.cfg.environment
    
    def parse(self, input_stream, apply_actions=True, pass_all=True, save_tokens=False, save_lines=False):
        if type(input_stream) is str or type(input_stream) is file:
            input_stream = self.lexer.tokenize(input_stream, save_tokens=save_tokens, save_lines=save_lines)
        self.parse_tree = self.get_parse_tree(input_stream)
        if apply_actions:
            return self.apply_actions(self.parse_tree)
        return self.parse_tree
        
        #~ success, derivation = self.get_derivation(input_stream)
        #~ if success:
            #~ self.parse_tree = self.build_parse_tree(derivation)
            #~ outcome = self.apply_actions(self.parse_tree, pass_all)
            #~ return outcome
        
    def get_parse_tree(self, input_stream):
        success = False
        eoi_sym = Terminal(EndOfInput())
        input_stream = iter(input_stream)
        input_stream = add_eoi(input_stream)
        root = None
        node_counter = 0
        stack = Stack([(self.cfg.start, root), (eoi_sym, None)])
        i = 0
        current = input_stream.next()
        try:
            while True:
                top, parent_node = stack.top()
                if top == eoi_sym:
                    if current == eoi_sym.value():
                        success = True
                        break
                    else:
                        raise ParsingException(self.lexer, current, "I wasn\'t expecting anything else, but there " \
                                                                    "still are some unread tokens.")
                elif isinstance(top, Terminal):
                    if current.matches(top.value()):
                        stack.pop()
                        new_node = ParseTreeNode(node_counter, Terminal(current))
                        node_counter += 1
                        if parent_node is not None: # and parent_node.symbol != Terminal(current):
                            parent_node.add_child(new_node)
                        elif parent_node is None:
                            root = new_node
                        current = input_stream.next()
                        i += 1
                    else:
                        raise ParsingException(self.lexer, current, "I was expecting \'%s\' but read \'%s\' instead." % (top.sym, current))
                elif isinstance(top, NonTerminal):
                    if current == eoi_sym.value():
                        rule_number = self.table[top, EmptyString.symbol]
                    else:
                        try:
                            rule_number = self.table[top, current.family()]
                        except KeyError:
                            raise ParsingException(self.lexer, current, "Illegal token.")
                    if rule_number is None:
                        firsts = '  '
                        full_first_list = list(self.table.non_terminal_first_sets[top])
                        firsts = ', '.join(str(token) for token in full_first_list)
                        raise ParsingException(self.lexer, current, "I was expecting one of the following tokens: %s. " \
                                                                    "(No rule to apply. Non-terminal symbol expected: %s.)" % (firsts,top.sym))
                    rule = self.cfg[rule_number]
                    new_node = ParseTreeNode(node_counter, top, rule_number)
                    node_counter += 1
                    if parent_node is not None:
                        parent_node.add_child(new_node)
                    else:
                        root = new_node
                    stack.pop()
                    reverse_rhs = reversed(rule.rhs)
                    for symbol in reverse_rhs:
                        stack.push((symbol, new_node))
                elif isinstance(top, EmptyString):
                    stack.pop()
                    new_node = ParseTreeNode(node_counter, EmptyString())
                    node_counter += 1
                    if parent_node is not None: # and parent_node.symbol != Terminal(current):
                        parent_node.add_child(new_node)
                    elif parent_node is None:
                        root = new_node
                else:
                    raise ParsingException(self.lexer, current, "There is an invalid element in the grammar: '"+str(top)+"'")
        except StopIteration:
            pass
        return root

    def get_derivation(self, input_stream):
        success = False
        eoi = EndOfInput()
        input_stream = iter(input_stream)
        input_stream = add_eoi(input_stream)
        stack = Stack([self.cfg.start, eoi])
        derivation = []
        i = 0
        current = input_stream.next()
        try:
            while True:
                top = stack.top()
                if top == eoi:
                    if current == eoi:
                        success = True
                        break
                    else:
                        (line,col) = position_in_text(input_stream, i)
                        raise ParsingException("I wasn't expecting anything else, but there still are some unread tokens, at position: "+str(i)+" line#"+str(line)+", col#"+str(col))
                elif isinstance(top, Terminal):
                    if top.value() == current:
                        stack.pop()
                        current = input_stream.next()
                        i += 1
                    else:
                        raise ParsingException("I was expecting '"+str(top)+"' but read '"+str(current)+"' instead.")
                elif isinstance(top, NonTerminal):
                    if current == eoi:
                        rule_number = self.table[top.value(), EmptyString.symbol]
                    elif current in self.cfg.terminals:
                        rule_number = self.table[top.value(), current]
                    else:
                        raise ParsingException("Illegal token: '" + str(current) + "' at position " + str(i) + "; context: " + str(current))
                    if rule_number is None:
                        raise ParsingException("No rule to apply")
                    rule = self.cfg[rule_number]
                    derivation.append(rule_number)
                    stack.pop()
                    reverse_rhs = reversed(rule.rhs)
                    for symbol in reverse_rhs:
                        stack.push(symbol)
                elif isinstance(top, EmptyString):
                    stack.pop()
                else:
                    raise ParsingException("There is an invalid element in the grammar: '"+str(top)+"'")
        except StopIteration:
            pass
        return (success, derivation)

    def build_parse_tree(self, rule_sequence, include_terminals=True):
        root = ParseTreeNode(0, self.cfg.start)
        counter = 0
        stack = Stack([(self.cfg.start, root)])
        for rule_id in rule_sequence:
            rule = self.cfg[rule_id]
            top_sym, top_node = stack.pop()
            while isinstance(top_sym, Terminal):
                top_sym, top_node = stack.pop()
            if top_sym != rule.lhs:
                raise ParsingException("Inconsistency: the stack's top and the rule's LHS do not match.")
            children = []
            reverse_rhs = reversed(rule.rhs)
            reverse_counter = counter + len(reverse_rhs)
            for sym in reverse_rhs:
                if include_terminals or isinstance(sym, NonTerminal):
                    new_node = ParseTreeNode(reverse_counter, sym)
                    stack.push((sym, new_node))
                    children = [new_node] + children
                    counter += 1
                    reverse_counter -= 1
            top_node.children = children
            top_node.rule_number = rule_id
        return root

    def apply_actions(self, node, pass_all=True):
        return self.__rec_apply_actions(node, pass_all)
    
    # Public mutators
    
    def __setstate__(self, dict):
        self.__dict__.update(dict)
        if type(self.rule_actions_module) is str:
            exec "import " + self.rule_actions_module
            self.rule_actions_module = eval(self.rule_actions_module)
    
    """
    Private instance methods
    """
    
    # Private accessors, selectors and extractors
    
    def __rec_apply_actions(self, node, pass_all=True, level=0):
        if node.isleaf():
            if isinstance(node.symbol, EmptyString):
                return node.symbol
            return node.symbol.value()
        rule = self.cfg[node.rule_number]
        children_outcome = []
        for child in node.children:
            if pass_all or isinstance(child.symbol, NonTerminal):
                child_outcome = self.__rec_apply_actions(child, pass_all, level + 2)
                if not isinstance(child_outcome, EmptyString):
                    children_outcome.append(child_outcome)
        if type(rule.action) is str:
            action_name = rule.action
        elif isinstance(rule.action, RuleAction):
            action_name = rule.action.name
        if self.rule_actions_module is None:
            action = rule.action
        elif type(self.rule_actions_module) is types.ModuleType:
            action = getattr(self.rule_actions_module, action_name)
        outcome = action(*children_outcome)
        return outcome


class ParsingTable:
    
    def __init__(self, cfg, discard_choices=False):
        log.info("Creating parsing table")
        self.cfg = cfg
        self.table = dict((n, dict((t.value(), set([])) \
                                   for t in cfg.terminals.union(SymbolSet([EmptyString()])))) \
                          for n in cfg.non_terminals)
        log.info("Filling table")
        self.__fill_table()
        log.info("Table filled")
        if discard_choices:
            log.info("Discarding choices")
            self.__discard_choices()
            log.info("Choices discarded")
        log.info("Parsing table created")
        self.col_size = 15
    
    def __setitem__(self, nt_pair, value):
        self.table[nt_pair[0]][nt_pair[1]] = value
        
    def __getitem__(self, nt_pair):
        return self.table[nt_pair[0]][nt_pair[1]]
    
    def __fill_table(self):
        first_sets = self.__compute_first_sets()
        self.first_sets = first_sets
        follow_sets = self.__compute_follow_sets(first_sets)
        self.follow_sets = follow_sets 
        non_terminal_first_sets, word_first_sets = first_sets
        self.non_terminal_first_sets = non_terminal_first_sets
        self.word_first_sets = word_first_sets
        rule_number = 0
        for rule in self.cfg.rules:
            non_terminal = rule.lhs
            #~ non_terminal = rule.lhs.value()
            first_set_has_the_empty_string = False
            for symbol in word_first_sets[rule_number]:
                #~ terminal = symbol
                terminal = symbol.value()
                self.table[non_terminal][terminal].add(rule_number)
                if isinstance(symbol, EmptyString):
                    first_set_has_the_empty_string = True
            if first_set_has_the_empty_string:
                for symbol in follow_sets[non_terminal]:
                    #~ terminal = symbol
                    terminal = symbol.value()
                    self.table[non_terminal][terminal].add(rule_number)
            rule_number += 1
    
    def __discard_choices(self):
        conflicts = 0
        for non_terminal in self.table:
            for terminal in self.table[non_terminal]:
                conflict = False
                this_set = self.table[non_terminal][terminal]
                if len(this_set) > 1:
                    conflicts += 1
                    conflict = True
                    print "Warning: the parser generator spotted a rule conflict; see log file."
                    log.warning("* Warning: the parser generator spotted a rule conflict:\n" + \
                                "  Rule conflict possible in the parsing table in the entry for \n" + \
                                "    Non-terminal: " + str(non_terminal) + "\n" \
                                "    Terminal:     " + str(terminal) + "\n" + \
                                "    Conflicting rules: " + str(this_set) + "\n")
                    for r in this_set:
                        log.warning("      " + str(r) + ": " + str(self.cfg.rules[r]))
                    log.warning("    Firsts set of " + str(non_terminal) + ": \n" + \
                                "      " + str([str(e) for e in self.first_sets[0][non_terminal]]) + "\n" + \
                                "    Follow set of " + str(non_terminal) + ": \n"
                                "      " + str([str(e) for e in self.follow_sets[non_terminal]]) + "\n")
                try:
                    selection = this_set.pop()
                    if conflict:
                        log.warning("  Choosing rule:\n" + \
                                    "    " + str(selection))
                    self.table[non_terminal][terminal] = selection
                except KeyError:
                    self.table[non_terminal][terminal] = None
        if conflicts == 0:
            print "* Parser successfully generated"
            log.info("Parser successfully generated")
        else:
            print "* Parser generated with conflicts"
            log.warning("Parser generated with conflicts")
    
    def __compute_first_sets(self):
        log.info("Computing first sets")
        non_terminal_first_sets = dict((n, set([])) for n in self.cfg.non_terminals)
        word_first_sets = [set([]) for rule in self.cfg.rules]
        first_sets = (non_terminal_first_sets, word_first_sets)
        previous_sets = (None, None)
        i = 0
        while first_sets != previous_sets:
            previous_sets = copy.deepcopy(first_sets)
            non_terminal_first_sets, word_first_sets = first_sets
            i+=1
            rule_number = 0
            for rule in self.cfg.rules:
                word_first_set = self.__get_word_first_set(rule.rhs, first_sets)
                word_first_sets[rule_number] |= word_first_set
                rule_number += 1
            rule_number = 0
            for rule in self.cfg.rules:
                non_terminal_first_sets[rule.lhs] |= word_first_sets[rule_number]
                rule_number += 1
            first_sets = (non_terminal_first_sets, word_first_sets)
        log.info("First sets computed")
        return first_sets

    
    def __get_word_first_set(self, word, first_sets):
        if len(word) == 0:
            return set([EmptyString()])
        elif isinstance(word[0], Terminal):
            return set([word[0]])
        elif isinstance(word[0], EmptyString) and len(word) == 1:
            return set([word[0]])
        non_terminal_first_sets, word_first_sets = first_sets
        non_terminal = word[0]
        this_set = non_terminal_first_sets[non_terminal]
        has_empty_string = False
        for element in this_set:
            if isinstance(element, EmptyString):
                has_empty_string = True
                break
        if not has_empty_string:
            return this_set
        else:
            empty_removed = set(element for element in this_set \
                                if not isinstance(element, EmptyString))
            rest = self.__get_word_first_set(word[1:], first_sets)
            result = empty_removed | rest
            return result
        
    def __get_word_first_set_with_empty_str(self, word, first_sets):
        if len(word) == 0:
            return set([])
        elif isinstance(word[0], Terminal):
            return set([word[0]])
        elif isinstance(word[0], EmptyString) and len(word) == 1:
            return set([word[0]])
        non_terminal_first_sets, word_first_sets = first_sets
        non_terminal = word[0]
        this_set = non_terminal_first_sets[non_terminal]
        has_empty_string = False
        for element in this_set:
            if isinstance(element, EmptyString):
                has_empty_string = True
                break
        if not has_empty_string:
            return this_set
        else:
            rest = self.__get_word_first_set_with_empty_str(word[1:], first_sets)
            result = this_set | rest
            return result
    
    def __compute_follow_sets(self, first_sets):
        log.info("Computing follow sets")
        follow_sets = dict((n, set([])) for n in self.cfg.non_terminals)
        previous_sets = None
        while follow_sets != previous_sets:
            previous_sets = copy.deepcopy(follow_sets)
            rule_number = 0
            for rule in self.cfg.rules:
                non_terminal = rule.lhs
                i = 0
                for symbol in rule.rhs:
                    if isinstance(symbol, NonTerminal):
                        rest = rule.rhs[i+1:]
                        followers = self.__get_word_first_set(rest, first_sets)
                        follow_sets[symbol] |= followers
                        followers_has_the_empty_string = False
                        for follower in followers:
                            if isinstance(follower, EmptyString):
                                followers_has_the_empty_string = True
                        if followers_has_the_empty_string or len(followers) == 0:
                            follow_sets[symbol] |= follow_sets[non_terminal]
                    i += 1
                rule_number += 1
        log.info("Follow sets computed")
        return follow_sets
    
    def __str__(self):
        s = " " * self.col_size + " "
        for t in self.cfg.terminals:
            token = t.value()
            if type(token) is types.ClassType:
                name = short_class_name(token)
            else:
                name = str(token)
            s += '|' + fit_str(name, self.col_size) + ' '
        s += '|' + str(EmptyString.symbol)
        s += '\n'
        s += '-' * (self.col_size + 1)
        s += ('+' + '-' * (self.col_size + 1)) * (len(self.cfg.terminals))
        s += '+' + '-' * max(self.col_size, len(str(EmptyString.symbol)))
        s += '\n'
        for n in self.cfg.non_terminals:
            s += fit_str(str(n.value()), self.col_size) + ' '
            for t in self.cfg.terminals:
                v = self.table[n][t.value()]
                if v is None:
                    v = '-'
                s += '|' + fit_str(str(v), self.col_size) + ' '
            last = self.table[n][EmptyString().value()]
            if last is None:
                last = '-'
            s += '|' + fit_str(str(last), self.col_size)
            s += '\n'
        return s


class ParseTreeNode:
    
    def __init__(self, identifier, symbol, rule_number=None, children=None):
        self.identifier = identifier
        self.symbol = symbol
        self.rule_number = rule_number
        if children is None:
            children = []
        self.children = children
        
    def isleaf(self):
        return self.children == []
    
    def add_child(self, node):
        self.children.append(node)
        
    def old_str(self):
        children_str = ""
        for child in self.children:
            lines = child.old_str().split('\n')
            for line in lines:
                children_str += " " * 4 + line + "\n"
        return "{\n" \
            +  "  id: " + str(self.identifier) + "\n" \
            +  "  symbol: '" + str(self.symbol) + "'\n" \
            +  "  rule: " + str(self.rule_number) + "\n" \
            +  "  children: \n" + children_str \
            +  "}\n"

    def __repr__(self):
        return "ParseTreeNode(identifier=%s, symbol=%s, rule_number=%s, children=%s)" % (self.identifier, self.symbol, self.rule_number, self.children)
    
    def __str__(self):
        return "{id: %s, symbol: %s, rule: %s, children: %d}" % (self.identifier, self.symbol, self.rule_number, len(self.children))

    def tree_str(self):
        s = str(self) + '\n'
        if len(self.children) > 0:
            s += '  |\n'
        lines = []
        i = 0
        while i < len(self.children):
            child = self.children[i]
            t = child.tree_str()
            lines = t.splitlines()
            if i < len(self.children) - 1:
                lines = ['  +- ' + lines[0]] \
                      + ['  |  ' + line for line in lines[1:]]
            else:
                lines = ['  +- ' + lines[0]] \
                      + ['     ' + line for line in lines[1:]]            
            i += 1
            s += '\n'.join(lines) + '\n'
        if len(self.children) == 0:
            return s + '\n'
        return s
    
    def pprint(self):
        print "{\n" \
            +  "  id: " + str(self.identifier) + "\n" \
            +  "  symbol: '" + str(self.symbol) + "'\n" \
            +  "  rule: " + str(self.rule_number) + "\n" \
            +  "  children: " + str([(n.identifier, str(n.symbol)) for n in self.children]) \
            +  "}\n"


class ParsingException(Exception):
    
    use_long_format = False
    short_format = "%(linenum)d:%(start)d-%(end)d:'%(substring)s':%(token)s:%(message)s"
    long_format  = "(Parsing exception)\n"\
                   "Line %(linenum)d; chars %(start)d-%(end)d:\n" \
                   "\n" \
                   "%(line)s\n" \
                   "%(underline)s\n" \
                   "\n" \
                   "Last token read: %(token)s\n" \
                   "%(message)s\n\n"
    
    def __init__(self, lexer, token, msg):
        self.token = token
        self.linenum = token.linenum
        self.start = token.start_col
        self.end = token.end_col
        if token.end_col is None:
            self.end_col = token.start_col
        self.msg = msg.rstrip().replace("'", "\\'")
        self.line = lexer.line.rstrip('\n\r')
    
    def __str__(self):
        if ParsingException.use_long_format:
            formatted_string = \
                ParsingException.long_format % { 'linenum':self.linenum,
                                                 'start':self.start,
                                                 'end':self.end,
                                                 'line':self.line,
                                                 'underline':" " * self.start + "^" * (self.end - self.start + 1),
                                                 'token':repr(self.token),
                                                 'message':self.msg }
        else:
            formatted_string = \
                ParsingException.short_format % { 'linenum':self.linenum,
                                                  'start':self.start,
                                                  'end':self.end,
                                                  'substring':self.line[self.start:self.end+1],
                                                  'token':repr(self.token),
                                                  'message':self.msg }
        return formatted_string

# Private functions

def add_eoi(input_stream):
    eoi = EndOfInput()
    if type(input_stream) is str:
        input_stream += eoi.value()
    elif type(input_stream) is list:
        input_stream += [eoi.value()]
    elif type(input_stream) is types.GeneratorType:
        def addend(generator, v):
            for item in generator:
                yield item
            yield v
        input_stream = addend(input_stream, eoi)
    return input_stream
    
def position_in_text(string, index):
    if index >= len(string):
        raise IndexError()
    i = 0
    line_counter = 0
    in_line = 0
    while i < index:
        char = string[i]
        if char is '\n':
            line_counter += 1
            in_line = 0
        i += 1
    return (line_counter, in_line)

def fit_str(s, n, filling=' '):
    if len(s) >= n:
        return s[:n]
    return s + filling * (n - len(s))


import pprint
pp = pprint.PrettyPrinter()


# Debugging functions

def print_first_sets(fs):
    ntfs, wfs = fs
    if fs == (None, None):
        print "Empty"
        return
    printable_ntfs = dict((str(n), set(str(s) for s in ntfs[n])) for n in ntfs)
    print "Non-terminal first-sets: "
    pp.pprint(printable_ntfs)
    print
    print_word_first_sets(wfs)
    print

def print_word_first_sets(wfs):
    if wfs == None:
        print "Wmpty"
        return
    printable_wfs = [[str(s) for s in w] for w in wfs]
    print "Word first_sets"
    pp.pprint(printable_wfs)

def print_word_first_set(s):
    printable_s = set(str(e) for e in s)
    pp.pprint(printable_s)
def print_follow_sets(fos):
    fos = dict((str(n), set(str(e) for e in fos[n])) for n in fos)
    pp.pprint(fos)

def window(input_stream, current, size=5):
    #~ return input_stream[current: current + size]
    return current
        

def print_stack(stack):
    print "  stack = "
    for s, n in stack:
        if n is not None:
            print '    ' + str(s) + ',' + str(n.identifier)
        else:
            print '    ' + str(s) + ',' + str(n)
    
