"""
Module: parser
Author: Ernesto Posse
Created on: June 22, 2006
Last update: Mar 16, 2007

Description:
  This module provides a factory function for creating parser objects, by giving the modules containing the CFG description and the rule actions
"""


import cPickle
import logging
import os.path
import sys

import aperiot
sys.path.append(aperiot.__path__[0])
sys.path.append(os.path.normpath(os.path.join(aperiot.__path__[0], '..')))

from aperiot.lexer import Lexer, Token
from aperiot.cfgs import CFG
from aperiot.llparser import Parser

logger = logging.getLogger(__name__)

# The speed of saving and loading parser objects depends on the pickle module.
# There are two alternative pickle modules in Python: pickle and cPickle
# The cPickle module is written in C and therefore is faster than the pickle
# module which is written in python itself. On the other hand, the cPickle module
# may suffer from some portability issues. If you want to improve the
# performance of saving and loading parsers, you can try uncommenting the following two lines...

pickle = cPickle

# ... and commenting out the following:

#~ import pickle




# The performance also depends on the pickle protocol (format) used. Currently
# There are 3 formats: 0, 1, and 2. Protocol 0 is an ASCII format which produces larger files and it takes longer to save and load, but is more portable. Protocols 1 ans 2 are binary formats which are faster to save and load but have some portability issues. Protocol 1 is meant for older versions of python. On the other hand, the ASCII protocol can be zipped with a very high compression rate.

ASCII      = 0
BINARY     = pickle.HIGHEST_PROTOCOL
OLD_BINARY = 1

formats = { ASCII: 'ASCII', BINARY: 'BINARY', OLD_BINARY: 'OLD_BINARY' }

#~ default_pickle_format = BINARY
default_pickle_format = ASCII

__all__ = ['ParserGenException', 'build_parser', 'save_parser', 'load_parser']


class ParserGenException(Exception):
    pass


def build_parser(cfg_description_module, lexer=None):
    logger.info("Building parser %s" % cfg_description_module)
    cfg_wat = cfg_description_module
    package_name = cfg_wat + '_cfg'
    full_module_name = package_name + '.' + cfg_wat
    try:
        exec "import " + full_module_name
    except NameError, e:
        raise ParserGenException(str(e))
    cfg_wat_module = eval(full_module_name)
    start = cfg_wat_module.start
    rules = cfg_wat_module.rules
    try:
        lexer_module = cfg_wat_module.lexer
        table = {}
        exec "import " + lexer_module in table
        for item in table:
            if issubclass(table[item], Lexer):
                lexer = table[item]
        if lexer is None:
            logger.warning("No subclass of aperiot.lexer.Lexer was found in module '%s'. Using default lexer instead." % lexer_module)
    except AttributeError:
        logger.info("Using default lexer")
    except ImportError:
        logger.error("Lexer module '%s' not found. Using default lexer instead." % lexer_module)
    useindents = cfg_wat_module.useindents
    usenewlines = cfg_wat_module.usenewlines
    modules_to_import = cfg_wat_module.imports
    tokens = [getattr(cfg_wat_module, object_name) 
                for object_name in dir(cfg_wat_module) 
                if isinstance(getattr(cfg_wat_module, object_name), Token)]
    if lexer is None:
        lexer = Lexer(tokens, 
                        yield_newlines=usenewlines, 
                        yield_indents=useindents)
    logger.info("Lexer created for " + cfg_description_module)
    cfg = CFG(start, rules, modules_to_import)
    logger.info("CFG created for " + cfg_description_module)
    logger.info("Normalizing CFG ... (this may take a few seconds or minutes if the grammar is large.)")
    cfg.normalize()
    logger.info("CFG normalized for " + cfg_description_module)
    parser = Parser(cfg, lexer, rule_actions_module=None)
    logger.info("Parser generated for " + cfg_description_module)
    return parser

def save_parser(parser, dir, module_name, pickle_format=default_pickle_format):
    logger.info("Saving parser %s in %s" % (module_name, dir))
    current_dir = os.getcwd()
    target_dir = os.path.abspath(dir)
    os.chdir(target_dir)
    root, _ = os.path.splitext(module_name)
    package_name = root + '_cfg'
    cfg = parser.cfg
    modules_to_import = cfg.modules_to_import
    if not cfg.normalized:
        cfg.save_package(package_name=package_name, 
                            module='llk', 
                            modules_to_import=modules_to_import, 
                            additional_actions=[])
    else:
        cfg.save_package(package_name=package_name, 
                            module='ll1', 
                            modules_to_import=modules_to_import, 
                            additional_actions=[])
    package_dir = os.path.join(target_dir, package_name)
    filename = os.path.join(package_dir, root + '.pkl')
    file_handler = file(filename, 'w')
    pickle.dump(parser, file_handler, pickle_format)
    abs_filename = os.path.abspath(filename)
    logger.info("Parser saved in file '" + abs_filename + "'")
    logger.info("File pickle format: " + formats[pickle_format])
    file_handler.close()
    os.chdir(current_dir)

def load_parser(module_name, package_name=None, path='.'):
    logger.info("Loading parser from module %s in package %s and with search path %s" % (module_name, package_name, path))
    sys.path.append(path)
    root, _ = os.path.splitext(module_name)
    if package_name is None:
        package_name = root + '_cfg'
    temp_table = {}
    exec "import " + package_name in temp_table
    package_path = eval(package_name, temp_table).__path__[0]
    filename = os.path.join(package_path, root + '.pkl')
    with file(filename, 'r') as file_handler:
        parser = pickle.load(file_handler)
        abs_filename = os.path.abspath(filename)
    parser.import_action_modules()
    logger.info("Parser loaded from file '"+abs_filename+"'")
    return parser


