"""
Module: plambda-ast
Author: Ernesto Posse

Description:
    Abstract syntax for plambda, a polyadic lambda-calcululs
"""

import types

__all__ = ['LambdaTerm', 'Constant', 'Variable', 'LambdaAbstraction', 'Application', 'SyntaxError', 'ExecutionError', 'ReductionError', 'Const', 'Var', 'Abs', 'App', 'LatentFunction']


class LambdaTerm:
    
    reductions = 0
    
    """
    Public instance methods
    """

    # Public instance accessors, selectors and extractors

    def execute(self, symbol_table={}):
        python_code = str(self)
        return eval(python_code, __builtins__, symbol_table)
        
    def msubst(self, mapping, symbol_table={}):
        t = self
        if type(mapping) is dict:
            mapping = mapping.items()
        for var, term in mapping:
            t = t.subst(var, term)
        return t
        

    def normalize(self, symbol_table={}, l=0):
        LambdaTerm.reductions += 1
        t = self
        try:
            t = self.__rec_normalize(symbol_table, l)
        except KeyError:
            pass
        return t
    

    """
    Private instance methods
    """
    
    # Private accessors, selectors and extractors
       
    
    # Private mutators
    
    def __rec_normalize(self, symbol_table={}, l=0):
        t = self
        if isinstance(t, Constant):
            pass
        elif isinstance(t, Variable):
            if t.name in symbol_table:
                t = symbol_table[t.name].term
                t = t.__rec_normalize(symbol_table, l+2)
        elif isinstance(t, LambdaAbstraction):
            new_body = t.body.__rec_normalize(symbol_table, l+2)
            t = LambdaAbstraction(t.params, new_body)
        elif isinstance(t, Application):
            new_func = t.function.__rec_normalize(symbol_table, l+4)
            new_args = [arg.__rec_normalize(symbol_table, l+4) for arg in t.args]
            t = Application(new_func, new_args)
            if isinstance(new_func, LambdaAbstraction) and not isinstance(new_func.body, Constant):
                t = t.beta_reduce()
                t = t.__rec_normalize(symbol_table, l+2)
        return t

    """
    Old methods
    """
    
    def deep_reduce(self, symbol_table={}, l=0):
        LambdaTerm.reductions += 1
        t = self
        try:
            t = self.__rec_deep_reduce(symbol_table, l)
        except KeyError:
            pass
        return t
    
    def deep_msubst(self, mapping, symbol_table={}):
        t = self
        for var, term in mapping:
            t = t.deep_subst(var, term)
        return t
    
    def deep_subst(self, var, term):
        return self.subst(var, term)
        
    def old_eval(self, symbol_table={}):
        LambdaTerm.reductions += 1
        t = self
        while isinstance(t, Application):
            if isinstance(t.function, Constant):
                break
            elif isinstance(t.function, Variable):
                t = Application(symbol_table[t.function.name].term, t.args)
            elif isinstance(t.function, LambdaAbstraction):
                t = t.beta_eval(symbol_table)
            elif isinstance(t.function, Application):
                t = Application(t.function.eval(symbol_table), t.args)
        return t
        
    def eval_msubst(self, mapping, symbol_table={}):
        t = self
        for var, term in mapping:
            t = t.subst(var, term)
        return t
        
    def get_value(self, symbol_table={}):
        return self.eval(symbol_table)
        
    def step(self):
        return self.beta_eval()

    def old_eval2(self):
        t = self
        while isinstance(t, Application):
            t = t.step()
        return t
    
    def __rec_deep_reduce(self, symbol_table={}, l=0):
        t = self
        if isinstance(t, Constant):
            pass
        elif isinstance(t, Variable):
            if t.name in symbol_table:
                t = symbol_table[t.name].term
                t = t.__rec_deep_reduce(symbol_table, l+2)
        elif isinstance(t, LambdaAbstraction):
            new_body = t.body.__rec_deep_reduce(symbol_table, l+2)
            t = LambdaAbstraction(t.params, new_body)
        elif isinstance(t, Application):
            new_func = t.function.__rec_deep_reduce(symbol_table, l+4)
            new_args = [arg.__rec_deep_reduce(symbol_table, l+4) for arg in t.args]
            t = Application(new_func, new_args)
            if isinstance(new_func, LambdaAbstraction):         # and not isinstance(new_func.body, Constant):
                t = t.beta_deep_reduce(symbol_table)
                t = t.__rec_deep_reduce(symbol_table, l+2)
        return t
        
    def __rec_reduce(self, symbol_table={}, l=0):
        t = self
        while isinstance(t, Application):
            if isinstance(t.function, Constant):
                break
            elif isinstance(t.function, Variable):
                t = Application(symbol_table[t.function.name].term, t.args)
            elif isinstance(t.function, LambdaAbstraction):
                if isinstance(t.function.body, Constant):
                    break
                t = t.beta_reduce(symbol_table)
            elif isinstance(t.function, Application):
                t = Application(t.function.__rec_reduce(symbol_table,l+2), t.args)
        return t
        
    def __str__(self):
        return '<plambda term>'
        
constants = []

class Constant(LambdaTerm):
    
    def __init__(self, value):
        global constants
        self.value = value
        constants += [self]
        
    def hasfree(self, variable):
        return False
        
    def freevars(self):
        return set()
        
    def subst(self, variable, term):
        return self
        
    def sim_subst(self, mapping):
        return self

    def evaluate(self, env,l=0):
        if type(self.value) is str \
        or isinstance(self.value, LatentFunction) \
        or isinstance(self.value, LambdaTerm):
            return eval(str(self.value), __builtins__, env)
        return self.value
        
    def __str__(self):
        return str(self.value)
        
    """
    Old methods
    """
    def deep_subst(self, variable, term):
        new_val = self.value.replace(variable, str(term))
        return Constant(new_val)
        
    def eval_msubst(self, mapping, symbol_table={}):
        env = {}
        for var, term in mapping:
            if isinstance(term, Constant):
                env[var] = term.value
            else:
                env[var] = term.get_value(symbol_table)
        env.update(symbol_table)
        result = eval(str(self.value), env)
        return result

        


class Variable(LambdaTerm):
    
    def __init__(self, name):
        self.name = name
        if type(name) is not str:
            raise SyntaxError('Bad variable name: "'+name+'" is not a string')
            
    def hasfree(self, variable):
        if variable == self.name:
            return True
        else:
            return False
            
    def freevars(self):
        return set([self.name])
        
    def subst(self, variable, term):
        if self.name == variable:
            return term
        else:
            return self
            
    def sim_subst(self, mapping):
        if self.name in mapping:
            return mapping[self.name]
        else:
            return self
            
    def evaluate(self, env,l=0):
        datum = env[self.name]
        return datum
        
    def __str__(self):
        return str(self.name)


class LambdaAbstraction(LambdaTerm):
    
    def __init__(self, params, body):
        self.params = params
        self.body = body
        if type(params) is not list:
            raise SyntaxError('Bad lambda abstraction: "'+str(params)+'" is not a valid parameter list.')
        for p in params:
            if type(p) is not str:
                raise SyntaxError('Bad lambda abstraction: parameter "'+str(p)+'" is not a valid parameter.')
        if not isinstance(body, LambdaTerm):
            raise SyntaxError('Bad lambda asbtraction: "'+str(body)+'" is not a valid plambda term')
            
    def hasfree(self, variable):
        if variable in self.params:
            return False
        return self.body.hasfree(variable)
        
    def freevars(self):
        return self.body.freevars() - set(self.params)
        
    def subst(self, variable, term):
        if variable in self.params or isinstance(self.body, Constant):
            return self
        else:
            common_vars = set(self.params) & term.freevars()
            if common_vars == set():
                return LambdaAbstraction(self.params, 
                                          self.body.subst(variable, term))
            else:
                new_params = []
                new_body = self.body
                for p in self.params:
                    if p in common_vars:
                        new_name = fresh_name()
                        new_params += [new_name]
                        new_body = new_body.subst(p, Var(new_name))
                    else:
                        new_params += [p]
                return LambdaAbstraction(new_params,
                                          new_body.subst(variable,term))
    
    def sim_subst(self, mapping):
        param_set = set(self.params)
        domain = set(mapping)
        if domain <= param_set:
            return self
        else:
            common_vars_domain = param_set & domain
            if common_vars_domain == set():
                my_free_vars = self.body.freevars()
                captured = []
                for y in param_set:
                    for x in domain:
                        if x in my_free_vars and y in mapping[x].freevars():
                            captured += [y]
                if len(captured) == 0:
                    return LambdaAbstraction(self.params, self.body.sim_subst(mapping))
                else:
                    new_params = []
                    renaming = {}
                    for p in self.params:
                        if p in captured:
                            new_name = fresh_name()
                            new_params += [new_name]
                            renaming[p] = Var(new_name)
                        else:
                            new_params += [p]
                    new_body = self.body.sim_subst(renaming)
                    return LambdaAbstraction(new_params, new_body.sim_subst(mapping))
            else:
                restricted_substitution = dict((var, mapping[var]) 
                                               for var in mapping if var not in common_vars_domain)
                return self.sim_subst(restricted_substitution)
               
    def evaluate(self, env,l=0):
        return LatentFunction(self, env)
        
    def __str__(self):
        params = ', '.join(self.params)
        return "lambda " + params + ": " + str(self.body)

    """
    Old methods
    """
    def deep_subst(self, variable, term):
        if variable in self.params:
            return self
        else:
            common_vars = set(self.params) & term.freevars()
            if common_vars == set():
                return LambdaAbstraction(self.params, 
                                          self.body.deep_subst(variable, term))
            else:
                new_params = []
                new_body = self.body
                for p in self.params:
                    if p in common_vars:
                        new_name = fresh_name()
                        new_params += [new_name]
                        new_body = new_body.deep_subst(p, Var(new_name))
                    else:
                        new_params += [p]
                return LambdaAbstraction(new_params,
                                          new_body.deep_subst(variable,term))
    

class Application(LambdaTerm):
    
    def __init__(self, function, args):
        self.function = function
        self.args = args
        if not isinstance(function, LambdaTerm):
            raise SyntaxError('Bad lambda application: "'+str(function)+'" is not a valid plambda term')
        if type(args) is not list and type(args) is not tuple:
            print str(type(args))
            raise SyntaxError('Bad lambda abstraction: "'+str(args)+'" is not a valid argument list.')
        new_args = []
        for p in args:
            if not isinstance(p, LambdaTerm) and not isinstance(p, LatentFunction):
                new_args += [Constant(p)]
            else:
                new_args += [p]
        self.args = new_args
    
    def hasfree(self, variable):
        return self.function.hasfree(variable) or any(arg.hasfree(variable) for arg in self.args)
        
    def freevars(self):
        args_free_vars = set()
        for arg in self.args:
            args_free_vars |= arg.freevars()
        return self.function.freevars() | args_free_vars
    
    def subst(self, variable, term):
        return Application(self.function.subst(variable, term), \
                            [arg.subst(variable, term) for arg in self.args])
    
    def sim_subst(self, mapping):
        return Application(self.function.sim_subst(mapping), \
                            [arg.sim_subst(mapping) for arg in self.args])
    
    def beta_reduce(self):
        new_subst = dict(zip(self.function.params, self.args))
        return self.function.body.sim_subst(new_subst)
        
    def evaluate(self, env, l=0):
        operator = self.function.evaluate(env,l+2)
        if isinstance(operator, LambdaAbstraction):
            operator = operator.evaluate(env,l+2)
        operands = []
        for arg in self.args:
            if isinstance(arg, LambdaTerm):
                operand = arg.evaluate(env, l+2)
            elif isinstance(arg, LatentFunction):
                operand = arg
            else:
                operand = Constant(arg).evaluate(env, l+2)
            operands += [operand]
        associations = dict(zip(operator.abstraction.params, operands))
        new_env = operator.env.extend(associations)
        
        if type(operator) is types.FunctionType:
            new_env['operator'] = operator
            new_env['operands'] = operands
            return eval('operator(*operands)', {}, new_env)
            
        if len(operator.abstraction.params) != len(self.args):
            raise ExecutionError('Wrong number of parameters. I was expecting: '+str(operator.abstraction.params)+' but received: '+str(self.args))
            
        result = operator.abstraction.body.evaluate(new_env,l+2)
        return result
    
    def __str__(self):
        args = ', '.join([str(arg) for arg in self.args])
        func = str(self.function) 
        if isinstance(self.function, LambdaAbstraction) \
        or isinstance(self.function, Application):
            func = "(" + func + ")"
        return func + "(" + args + ")"

    """
    Old methods
    """
    def beta_deep_reduce(self, symbol_table={}):
        new_subst = zip(self.function.params, self.args)
        return self.function.body.deep_msubst(new_subst, symbol_table)
        
    def beta_eval(self, symbol_table={}):
        new_subst = zip(self.function.params, self.args)
        return self.function.body.eval_msubst(new_subst, symbol_table)
    
    def deep_subst(self, variable, term):
        return Application(self.function.deep_subst(variable, term), \
                            [arg.deep_subst(variable, term) for arg in self.args])
    
    def old_step(self):
        if isinstance(self.function, Constant):
            return self
        elif isinstance(self.function, Variable):
            return self
        elif isinstance(self.function, LambdaAbstraction):
            return self.beta_eval()
        elif isinstance(self.function, Application):
            return Application(self.function.step(), self.args)
    

class SyntaxError(Exception):
    pass


class ExecutionError(Exception):
    pass


class ReductionError(Exception):
    pass


Const = Constant
Var = Variable
Abs = LambdaAbstraction
App = Application


class LatentFunction:

    def __init__(self, abstraction, env):
        self.abstraction = abstraction
        self.env = env
        
    def __str__(self):
        return str(self.abstraction)


name_counter = 0

def fresh_name():
    global name_counter
    name_counter += 1
    return 'z'+str(name_counter)
